program DecimalOperatorTest;

{$APPTYPE CONSOLE}

uses
  SysUtils,
  Math,
  Decimals in '..\Decimals.pas';

type
  TTestResultInfo = (
    triOk              = 0,
    triOverflow        = 1,
    triReverseRound    = 2,
    triDivideByZero    = 3,
    triReverseOverflow = 4
  );

  TTestResult = record
    info: TTestResultInfo;
    val: string;
    constructor Create(i: TTestResultInfo; const v: string);
  end;

{$I 'DecimalTestData.inc'}

procedure WriteFmt(const Fmt: string; Args: array of const); overload;
begin
  Writeln(Format(Fmt, Args));
end;

const
  Bools: array[Boolean] of string = ('FALSE', 'TRUE');

function Dump(const D: Decimal): string;
var
  L: TLongwordArray;
  S: string;
begin
  L := Decimal.GetBits(D);
  if (L[3] and $80000000) <> 0 then S := '-' else S := '+';
  Result := Format('%s %.2x %.8x %.8x %.8x', [S, L[3] shr 16 and $1F, L[2], L[1], L[0]]);
end;

procedure Error(const msg: string; i, j: Integer; d1, d2, d3, d3b: Decimal);
var
  delta: Decimal;
begin
  delta := 0;
  try
    delta := d3 - d3b;
  except
    on E: Exception do
      WriteFmt('Error: unexpected exception on %s - %s; %s', [d3.ToString, d3b.ToString, e.Message]);
  end;
  Writeln;
  WriteFmt('(%d,%d): %s', [i, j, msg]);
  Writeln;
  WriteFmt('d1:    %31s  %s', [d1.ToString, Dump(d1)]);
  WriteFmt('d2:    %31s  %s', [d2.ToString, Dump(d2)]);
  WriteFmt('act:   %31s  %s', [d3.ToString, Dump(d3)]);
  WriteFmt('exp:   %31s  %s', [d3b.ToString, Dump(d3b)]);
  WriteFmt('delta: %31s  %s', [delta.ToString, Dump(delta)]);
  Writeln;
end;

procedure TestComp;
var
  i: Integer;
  j: Integer;
  x: TValueSign;
  y: TValueSign;
  d1: Decimal;
  d2: Decimal;
  b: Boolean;
  count: Integer;
begin
  Writeln('Comparison');
  Writeln;
  count := Length(CompArguments);

  for i := 0 to count - 1 do
  begin
    d1 := CompArguments[i];
    for j := 0 to count - 1 do
    begin
      if (j mod (CompCount div 2)) = 0 then
        Writeln;

      Assert(CompResults[i, j] = -CompResults[j, i]);
      x := CompResults[i, j];
      d2 := CompArguments[j];

      Write(Format('(%d,%d)', [i, j]):10);

      y := Decimal.Compare(d1, d2);
      Assert(x = y);

      y := d1.CompareTo(d2);
      Assert(x = y);

      b := d1 < d2;
      if x <> -1 then
        b := not b;
      Assert(b);

      b := d1 <= d2;
      if x = 1 then
        b := not b;
      Assert(b);

      b := d1 >= d2;
      if x = -1 then
        b := not b;
      Assert(b);

      b := d1 > d2;
      if x <> 1 then
        b := not b;
      Assert(b);

      b := d1 = d2;
      if x <> 0 then
        b := not b;
      Assert(b);

      b := d1.Equals(d2);
      if x <> 0 then
        b := not b;
      Assert(b);

      b := Decimal.Equals(d1, d2);
      if x <> 0 then
        b := not b;
      Assert(b);

    end;
  end;
  Writeln;
  Writeln;
end;

procedure TestAdd;
var
  errOverflow: Integer;
  errOp: Integer;
  count: Integer;
  n: Integer;
  d1: Decimal;
  d2: Decimal;
  d3: Decimal;
  d4: Decimal;
  i: Integer;
  j: Integer;
  tr: TTestResult;
begin
  errOverflow := 0;
  errOp := 0;
  count := Length(Arguments);
  n := 0;

  Writeln('Addition');
  Writeln;
  for i := 0 to count - 1 do
  begin
    d1 := Arguments[i];
    for j := 0 to count - 1 do
    begin
      d2 := Arguments[j];
      d3 := 0;
      d4 := 0;
      tr := AddResults[n];
      try
        Write(Format('(%d,%d)', [i, j]):10);
        d3 := Decimal.Add(d1, d2);
        if d3 <> tr.val then
          if tr.info = triOverflow then
          begin
            Error('Add: expected overflow', i, j, d1, d2, d3, tr.val);
            Inc(errOverflow);
          end
          else
          begin
            Error('Add: result mismatch', i, j, d1, d2, d3, tr.val);
            Inc(errOp);
          end
        else if tr.info = triOk then
        begin
          d4 := Decimal.Subtract(d3, d2);
          if d4 <> d1 then
          begin
            Error('Subtract: result mismatch', i, j, d3, d2, d4, d1);
            Inc(errOp);
          end;
        end;
      except
        on E: EOverflow do
        begin
          if tr.info <> triOverflow then
          begin
            Error('Add: unexpected overflow', i, j, d1, d2, d3, 0);
            Inc(errOverflow);
          end;
        end;
      end;
      Inc(n);
    end;
  end;
  Writeln;
  Writeln(errOp, ' wrong additions, ', errOverflow, ' wrong overflows');
  Writeln;
end;

procedure TestMult;
var
  errOverflow: Integer;
  errOp: Integer;
  count: Integer;
  n: Integer;
  i: Integer;
  j: Integer;
  d1: Decimal;
  d2: Decimal;
  d3: Decimal;
  d4: Decimal;
  tr: TTestResult;
begin
  errOverflow := 0;
  errOp := 0;
  count := Length(Arguments);
  n := 0;

  Writeln('Multiplication');
  Writeln;
  for i := 0 to count - 1 do
  begin
    d1 := Arguments[i];
    for j := 0 to count - 1 do
    begin
      d2 := Arguments[j];
      d3 := 0;
      d4 := 0;
      tr := MultResults[n];
      try
        Write(Format('(%d,%d)', [i, j]):10);
        d3 := Decimal.Multiply(d1, d2);
        if d3 <> tr.val then
        begin
          if tr.info = triOverflow then
          begin
            Error('Mult: expected overflow', i, j, d1, d2, d3, tr.val);
            Inc(errOverflow);
          end
          else
          begin
            Error('Mult: result mismatch', i, j, d1, d2, d3, tr.val);
            Inc(ErrOp);
          end;
        end;
      except
        on E: EOverflow do
        begin
          if tr.info <> triOverflow then
          begin
            Error('Mult: unexpected overflow', i, j, d1, d2, d3, tr.val);
            Inc(errOverflow);
          end;
        end;
      end;

      if (d2 <> 0) and (tr.info <> triOverflow) then
        try
          d4 := Decimal.Divide(d3, d2);
          if (d4 <> d1) and (tr.info <> triReverseRound) then
          begin
            Error('MultDiv: result mismatch', i, j, d3, d2, d4, d1);
            Inc(errOp);
          end;
        except
          on E: Exception do
            if tr.info <> triReverseOverflow then
            begin
              Error('MultDiv: unexpected overflow', i, j, d3, d2, d4, d1);
              Inc(ErrOverflow);
            end;
        end;
      Inc(n);
    end;
  end;
  Writeln;
  Writeln(errOp, ' wrong multiplications, ', errOverflow, ' wrong overflows');
  Writeln;
end;

procedure TestDiv;
var
  errOverflow: Integer;
  errDividebyZero: Integer;
  errOp: Integer;
  count: Integer;
  n: Integer;
  i: Integer;
  j: Integer;
  d1: Decimal;
  d2: Decimal;
  d3: Decimal;
  d4: Decimal;
  tr: TTestResult;
begin
  errOverflow := 0;
  errDivideByZero := 0;
  errOp := 0;
  count := Length(Arguments);
  n := 0;

  Writeln('Division');
  Writeln;
  for i := 0 to count - 1 do
  begin
    d1 := Arguments[i];
    for j := 0 to count - 1 do
    begin
      d2 := Arguments[j];
      d3 := 0;
      d4 := 0;
      tr := DivResults[n];
      try
        Write(Format('(%d,%d)', [i, j]):10);
        d3 := Decimal.Divide(d1, d2);
        if d3 <> tr.val then
        case tr.info of
          triOverflow:
          begin
            Error('Div: expected overflow', i, j, d1, d2, d3, tr.val);
            Inc(errOverflow);
          end;
          triDivideByZero:
          begin
            Error('Div: expected divide by zero', i, j, d1, d2, d3, tr.val);
            Inc(errDivideByZero);
          end;
          else
          begin
            Error('Div: result mismatch', i, j, d1, d2, d3, tr.val);
            Inc(errOp);
          end;
        end;
      except
        on E: EOverflow do
          if tr.info <> triOverflow then
          begin
            Error('Div: unexpected overflow', i, j, d1, d2, d3, 0);
            Inc(errOverflow);
          end;
        on E: EZeroDivide do
          if tr.info <> triDividebyZero then
          begin
            Error('Div: unexpected division by zero', i, j, d1, d2, d3, 0);
            Inc(errDivideByZero);
          end;
      end;

      if d3 <> Decimal.Zero then
        try
          d4 := Decimal.Multiply(d3, d2);
          if (d4 <> d1) and (tr.info <> triReverseRound) then
          begin
            Error('DivMult: result mismatch', i, j, d3, d2, d4, d1);
            Inc(errOp);
          end;
        except
          on E: EOverflow do
          begin
            if tr.info <> triReverseOverflow then
            begin
              Error('DivMult: unexpected overflow', i, j, d3, d2, d4, d1);
              Inc(errOverflow);
            end;
          end;

        end;
      Inc(n);
    end;
  end;
  Writeln;
  Writeln(errOp, ' wrong divisions, ', errOverflow, ' wrong overflows, ', errDivideByZero, ' wrong divisions by zero');
  Writeln;
end;

procedure TestRem;
var
  errOverflow: Integer;
  errDividebyZero: Integer;
  errOp: Integer;
  count: Integer;
  n: Integer;
  i: Integer;
  j: Integer;
  d1: Decimal;
  d2: Decimal;
  d3: Decimal;
  d4: Decimal;
  tr: TTestResult;
begin
  errOverflow := 0;
  errDivideByZero := 0;
  errOp := 0;
  count := Length(Arguments);
  n := 0;

  Writeln('Remainder');
  Writeln;
  for i := 0 to count - 1 do
  begin
    d1 := Arguments[i];
    for j := 0 to count - 1 do
    begin
      d2 := Arguments[j];
      d3 := 0;
      d4 := 0;
      tr := RemResults[n];
      try
        Write(Format('(%d,%d)', [i, j]):10);
        d3 := Decimal.Remainder(d1, d2);
        if d3 <> tr.val then
        case tr.info of
          triOverflow:
          begin
            Error('Rem: expected overflow', i, j, d1, d2, d3, tr.val);
            Inc(errOverflow);
          end;
          triDivideByZero:
          begin
            Error('Rem: expected divide by zero', i, j, d1, d2, d3, tr.val);
            Inc(errDivideByZero);
          end;
          else
          begin
            Error('Rem: result mismatch', i, j, d1, d2, d3, tr.val);
            Inc(errOp);
          end;
        end;
      except
        on E: EOverflow do
          if tr.info <> triOverflow then
          begin
            Error('Rem: unexpected overflow', i, j, d1, d2, d3, 0);
            Inc(errOverflow);
          end;
        on E: EZeroDivide do
          if tr.info <> triDividebyZero then
          begin
            Error('Rem: unexpected division by zero', i, j, d1, d2, d3, 0);
            Inc(errDivideByZero);
          end;
      end;

      if (d3 <> Decimal.Zero) and (d2 <> Decimal.Zero) then
        try
          d4 := d3 + Decimal.Truncate(d1 / d2) * d2;
          if (d4 <> d1) and (tr.info <> triReverseRound) then
          begin
            Error('RemDivMult: result mismatch', i, j, d3, d2, d4, d1);
            Inc(errOp);
          end;
        except
          on E: EOverflow do
          begin
            if tr.info <> triReverseOverflow then
            begin
              Error('RemDivMult: unexpected overflow', i, j, d3, d2, d4, d1);
              Inc(errOverflow);
            end;
          end;

        end;
      Inc(n);
    end;
  end;
  Writeln;
  Writeln(errOp, ' wrong remainders, ', errOverflow, ' wrong overflows, ', errDivideByZero, ' wrong divisions by zero');
  Writeln;
end;

procedure DumpArguments;
var
  I: Integer;
  D: Decimal;
begin
  Writeln('Arguments');
  Writeln;
  for I := Low(Arguments) to High(Arguments) do
  begin
    D := Arguments[I];
    WriteFmt('%.2d: %31s   %s', [I, D.ToString, Dump(D)]);
  end;
  Writeln;
end;

{ TTestResult }

constructor TTestResult.Create(i: TTestResultInfo; const v: string);
begin
  info := i;
  val := v;
end;

begin
  try
    DumpArguments;
    TestComp;
    TestRem;
    TestAdd;
    TestMult;
    TestDiv;
  except
    on E: Exception do
      Writeln(E.ClassName, ': ', E.Message);
  end;
  if DebugHook <> 0 then
  begin
    Writeln;
    Write('Press any key...');
    Readln;
  end;
end.

